#!/bin/env python

import yaml
import sys
import logging
import argparse
from pprint import pprint
from rocolib.library import getComponent, getComponentTree
from rocolib.api.composables.graph.Joint import FingerJoint


def test(component, params, thickness, display=False, display3D=False):
    f = getComponent(component)
    if params is not None:
        for p in params:
            try:
                v = eval(str(p[1]))
            except NameError:
                # Using a string as a dimensions parameter
                v = str(p[1])
            f.setParameter(p[0], v)
    if thickness is None:
        t = 0
        j = None
    else:
        t = thickness
        j = FingerJoint(thickness=t)

    f.makeOutput("output/" + component, display=display, display3D=display3D, thickness=t, joint=j)

def cli_argparse():
    LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    DEFAULT_LOG_LEVEL = "WARNING"

    parser = argparse.ArgumentParser()

    parser.add_argument("component", nargs='?',
                        help="Name of the component you'd like to test")
    parser.add_argument("--file", "-f", type=str, help="Load component and parameters from file")
    parser.add_argument("--list", "-l", action='store_true',
                        help="List all known library components")

    parser.add_argument("--verbose", "-v", help="Increase logging level (can be used multiple times)",
                        dest="log_level", action="append_const", const=-1)
    parser.add_argument("--quiet", "-q", help="Decrease logging level (can be used multiple times)",
                        dest="log_level", action="append_const", const=1)
    parser.add_argument("-I", help="List component interfaces (use multiple times for more detail)",
                        dest="interface_list", action="append_const", const=1)
    parser.add_argument("-P", help="List component parameters (use multiple times for more detail)",
                        dest="param_list", action="append_const", const=1)
    parser.add_argument("-t", type=float, help="Thickness (i.e. making with wood)")
    parser.add_argument("-d", action='store_true', help="Display 2D drawing")
    parser.add_argument("-D", action='store_true', help="Display 3D drawing")
    parser.add_argument("-p", nargs=2, action='append',
        metavar=("NAME", "VALUE"),
        help="Set component parameter NAME to value VALUE")

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    log_level = LOG_LEVELS.index(DEFAULT_LOG_LEVEL)
    # For each "-q" and "-v" flag, adjust the logging verbosity accordingly
    # making sure to clamp off the value from 0 to 4, inclusive of both
    for adjustment in args.log_level or ():
        log_level = min(len(LOG_LEVELS) - 1, max(log_level + adjustment, 0))
    log_level_name = LOG_LEVELS[log_level]
    logging.basicConfig(level=getattr(logging, log_level_name))

    acted = False
    if args.list:
        t = getComponentTree()
        for i, cs in enumerate(t):
            chunks = [cs[x:x+4] for x in range(0, len(cs), 4)]
            print(f"{i:2}: " + "\n    ".join((" ".join(x) for x in chunks)))
        acted = True

    if args.file:
        with open(args.file) as f:
            config = yaml.safe_load(f)
        args.component = args.component or config.pop("$component", None)
        args.t = args.t or config.pop("$thickness", None)
        if args.p:
            config.update(dict(args.p))
        args.p = list(config.items())

    if args.component:
        if args.interface_list:
            f = getComponent(args.component)
            info = f.getInterfaceInfo(len(args.interface_list))
            if len(args.interface_list) == 1:
                print("\n".join(info))
            else:
                pprint(info)

            if not args.p or args.t or args.d or args.D:
                exit(0)
        if args.param_list:
            f = getComponent(args.component)
            info = f.getParameterInfo()
            if len(args.param_list) == 1:
                print("\n".join(info.keys()))
            elif len(args.param_list) == 2:
                pprint({k: v['defaultValue'] for k,v in info.items()})
            elif len(args.param_list) == 3:
                pprint({k: {x: v[x] for x in ('defaultValue', 'spec')} for k,v in info.items()})
            else:
                pprint(info)

            if not args.p or args.t or args.d or args.D:
                exit(0)
        test(args.component, args.p, thickness=args.t, display=args.d, display3D=args.D)
        acted = True

    if not acted:
        parser.print_help(sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    cli_argparse()