from collections import OrderedDict
import os, posixpath
import sys
import yaml
import logging
import networkx as nx
from circlify import circlify

from rocolib import ROCOLIB_DIR
from rocolib.api.Parameterized import Parameterized
from rocolib.api.Function import YamlFunction
from rocolib.utils.utils import prefix as prefixString
from rocolib.utils.utils import tryImport
from rocolib.utils.io import load_yaml
from rocolib.utils.nx2go import GraphVisualization as gv


log = logging.getLogger(__name__)
# XXX circular import: 
# from rocolib.library import ROCOLIB_LIBRARY
# instead: 
ROCOLIB_LIBRARY = os.path.join(ROCOLIB_DIR, "library")

def getSubcomponentObject(component, name=None):
    try:
        obj = tryImport(component, component)
        # XXX hack to get around derived components not having name parameter in their __init__
        c = obj()
        c.setName(name)
        return c
    except ImportError:
        c = Component(component)
        c.setName(name)
        return c

class Component(Parameterized):
    @classmethod
    def test(cls, params=None, **kwargs):
        c = cls()
        if params:
            for key, val in params.items():
                c.setParameter(key, val)
        filedir = kwargs.pop('filedir', None)
        c.makeOutput(filedir, **kwargs)

    def __init__(self, yamlFile=None):
        self._name = None
        Parameterized.__init__(self)
        self.reset()
        yf = yamlFile
        if not yamlFile:
            yf = type(self).__name__ + ".yaml"
        for fn in (yf, 
                   os.path.join(ROCOLIB_DIR, yf), 
                   os.path.join(ROCOLIB_LIBRARY, yf)):
            try:
                self.fromYaml(fn)
                break
            except IOError:
                pass
            if yamlFile:
                raise ValueError(f"No suitable yamlfile found for {yamlFile}")
        self.predefine()
        self.define()

    def getName(self):
      return self._name if self._name is not None else str(self.__class__)

    def setName(self, name):
      self._name = name

    def fromYaml(self, filename):
        definition = load_yaml(filename)

        # keys are (parameters, subcomponents, connections, interfaces)
        if "parameters" in definition:
          for k, v in definition["parameters"].items():
              self.addParameter(k, v["defaultValue"], **v["spec"])

        if "subcomponents" in definition:
          val = definition["subcomponents"]
          for k, v in val.items():
            self.addSubcomponent(k, v["classname"], **v["kwargs"])
            self.subcomponents[k]["parameters"] = v["parameters"]

        if "connections" in definition:
          self.connections = definition["connections"]

        if "interfaces" in definition:
          val = definition["interfaces"]
          for k, v in val.items():
            if isinstance(v, dict):
              self.inheritInterface(k, (v["subcomponent"], v["interface"]))
            else:
              self.interfaces[k] = v
          self.reinheritAllInterfaces()

    def reset(self):
        # Used during design
        self.parameters = {}
        self.subcomponents = {}
        self.connections = {}
        self.interfaces = {}
        self.defaults = {}
        self.semanticConstraints = []

        # Used during make
        self.composables = OrderedDict()

    def predefine(self, **kwargs):
        ### Override in Component subclass to define subclass parameters, interfaces, etc.
        pass

    def define(self):
        ### Override in Component instance to define individual parameters, interfaces, etc.
        pass

    def getInterfaceInfo(self, detail=1):
        if detail == 1:
            return list(self.interfaces.keys())

        if detail == 2:
            return {x: self.getInterface(x, transient=True).__class__.__name__ for x in self.interfaces}

        i =  {x: dict(port = self.getInterface(x, transient=True).__class__.__name__) for x in self.interfaces}
        for k, v in self.interfaces.items():
            if isinstance(v, dict):
                i[k].update(v)
        return i

    ###
    # DESIGN PHASE
    ###

    def addSubcomponent(self, name, classname, inherit=False, prefix = "", **kwargs):
        '''

        :param name: unique identifier to refer to this component by
        :type  name: str or unicode
        :param classname: code name of the subcomponent
                    should be python file/class or yaml name
        :type  classname: str or unicode
        '''
        # XXX will silently fail if subcomponent name is already taken?
        obj = getSubcomponentObject(classname, self.getName() + '.' + name)
        sc = {"classname": classname, "parameters": {}, "kwargs": kwargs, "object": obj}
        self.subcomponents.setdefault(name, sc)

        if inherit:
            if prefix == "":
                prefix = name

            for key, value in obj.getParameterInfo().items():
                # inherit = True : inherit all parameters
                if inherit is True or key in inherit:
                    try:
                        self.addParameter(prefixString(prefix, key), value["defaultValue"], **value["spec"])
                    except KeyError:
                        # It's ok if we try to add a parameter that already exists
                        pass
                    self.addConstraint((name, key), prefixString(prefix, key))
            # XXX also inherit interfaces?

        return self

    def delSubcomponent(self, name):
        toDelete = []

        # delete edges connecting components
        for connName, ((fromComp, _), (toComp, _), _) in self.connections.items():
            if name in (fromComp, toComp):
                toDelete.append(connName)
        for connName in toDelete:
            self.connections.pop(connName)

        self.subcomponents.pop(name)

    def addConstraint(self, constraint_params, inputs, function=None, subcomponent=None):
        data = {"parameter": inputs}
        if subcomponent:
            data["subcomponent"] = subcomponent
        if function:
            data["function"] = function

        self.addConstConstraint(constraint_params, data)

    def addConstConstraint(self, constraint_params, value):
        (subComponent, parameterName) = constraint_params

        # If constraint exists and was inherited
        # i.e. is the constant function for prefixstring'ed 
        # then remove the inherited parameter 
        newkey = prefixString(subComponent, parameterName)
        try:
            if self.hasParameter(newkey) \
                and self.getSubParameters(subComponent).get(parameterName).get("parameter") == newkey \
                and self.getSubParameters(subComponent).get(parameterName).get("function") is None:
                    if self.getParameterInfo()[newkey]["spec"].get("optional"):
                        for override in self.getParameterInfo()[newkey]["spec"].get("overrides", ()):
                            self.delParameter(prefixString(subComponent, override))
                            self.delConstraint(subComponent, override)
                    self.delParameter(newkey)
        except AttributeError as e:
            pass

        # XXX otherwise silently overwrites existing constraints, is that ok?
        self.getSubParameters(subComponent)[parameterName] = value

    def delConstraint(self, subComponent, parameterName):
        self.getSubParameters(subComponent).pop(parameterName)

    def addInterface(self, name, val):
        if name in self.interfaces:
            raise ValueError("Interface %s already exists" % name)
        self.interfaces.setdefault(name, val)
        return self

    def inheritAllInterfaces(self, subcomponent, prefix=""):
        obj = self.subcomponents[subcomponent]["object"]
        if prefix == "":
          prefix = subcomponent
        for name, value in obj.interfaces.items():
          self.inheritInterface(prefixString(prefix, name), (subcomponent, name))
        self.reinheritAllInterfaces()
        return self

    def inheritInterface(self, name, interface_params):
        (subcomponent, subname) = interface_params
        if name in self.interfaces:
            raise ValueError("Interface %s already exists" % name)
        value = {"subcomponent": subcomponent, "interface": subname}
        obj = self.subcomponents[subcomponent]["object"]
        iobj = obj.getInterface(subname, transient=True).inherit(self, subcomponent)
        if iobj:
          value["object"] = iobj
        self.interfaces.setdefault(name, value)

        return self

    # XXX kinda hackish?
    def reinheritAllInterfaces(self):
      for name, value in self.interfaces.items():
        if isinstance(value, dict):
          subc = value["subcomponent"]
          subi = value["interface"]
          obj = self.subcomponents[subc]["object"]
          iobj = obj.getInterface(subi, transient=True).inherit(self, subc)
          if iobj:
            value["object"] = iobj
          self.interfaces.setdefault(name, value)

    # Create both constraints and connections
    def join(self, fromInterface, toInterface, name=None, **kwargs):
        fromComponent = fromInterface[0]
        fromPort = self.getInterfaces(*fromInterface)
        toComponent = toInterface[0]
        toPort = self.getInterfaces(*toInterface)
        for fromParam, toParam in zip(fromPort.getParams(), toPort.getParams()):
            self.addConstraint((toComponent, toParam), fromParam, subcomponent=fromComponent)
        self.addConnection(fromInterface, toInterface, name, **kwargs)

    def addConnection(self, fromInterface, toInterface, name=None, **kwargs):
        if name is None:
          for i in range(len(self.connections)+1):
            name = "connection%d" % i
            if name not in self.connections:
              break
        for k, v in kwargs.items():
          try:
            kwargs[k] = v.toYamlObject()
          except AttributeError:
            pass

        fromPort = self.getInterfaces(*fromInterface)
        toPort = self.getInterfaces(*toInterface)
        if fromPort.canMate(toPort) or toPort.canMate(fromPort):
            self.connections.setdefault(name, [fromInterface, toInterface, kwargs])
        else:
            raise AttributeError(f"{fromInterface} ({fromPort} : {type(fromPort)}) cannot connect to {toInterface} ({toPort} : {type(toPort)}) according to getMate")

    def delConnection(self, name):
      self.connections.pop(name)

    def getConnections(self, component1, component2=None):
      keys = []
      for connName, (fromInterface, toInterface, kwargs) in self.connections.items():
        if component1 in [fromInterface[0], toInterface[0]]:
          if component2 is None or component2 in [fromInterface[0], toInterface[0]]:
            keys.append(connName)
      return keys
          

    '''
    # TODO : delete Interface
    # XXX : remove constraints that involve this parameter?
    def delParameter(self, name):
    '''

    def toLibrary(self, name):
        # XXX TODO: Check for collisions!  
        # if collision:
        #   flag to allow rebuilding, and fail otherwise?  
        #   if no flag, check if source matches and rebuild if so, fail otherwise?
        return self.toYaml(ROCOLIB_LIBRARY, name + ".yaml")

    def toYaml(self, basedir, filename):
        filepath = os.path.join(basedir, filename)
        source = os.path.relpath(sys.argv[0], basedir).replace(os.sep, posixpath.sep)

        parameters = {}
        for k, v in self.parameters.items():
          parameters[k] = {"defaultValue": v.defaultValue, "spec": v.spec}

        subcomponents = {}
        for k, v in self.subcomponents.items():
          subcomponents[k] = {"classname": v["classname"], "parameters": v["parameters"], "kwargs": v["kwargs"]}

        interfaces = {}
        for k, v in self.interfaces.items():
          if isinstance(v, dict):
            interfaces[k] = {"subcomponent": v["subcomponent"], "interface": v["interface"]}
          else:
            interfaces[k] = v

        definition = {
            "source" : source,
            "parameters" : parameters,
            "subcomponents" : subcomponents,
            "connections" : self.connections,
            "interfaces" : interfaces,
        }

        with open(filepath, "w") as fd:
            yaml.safe_dump(definition, fd)

    ###
    # GETTERS AND SETTERS
    ###

    def getSubcomponent(self, name):
        return self.subcomponents[name]['object']

    def getSubParameters(self, name):
        return self.subcomponents[name]['parameters']

    def setSubParameter(self, c, n, v):
        self.getSubcomponent(c).setParameter(n, v)

    def getInterfaces(self, component, name, transient=False):
        return self.getSubcomponent(component).getInterface(name, transient)

    def getInterface(self, name, transient=False):
        c = self.interfaces[name]

        if isinstance(c, dict):
            if "object" in c and c["object"]: 
              return c["object"]
            
            subc = c["subcomponent"]
            subi = c["interface"]
            i = self.getInterfaces(subc, subi, transient)
            if not transient:
              c["object"] = i
            return i
        else:
          return c

    def setInterface(self, n, v):
        if n in self.interfaces:
            self.interfaces[n] = v
        else:
            raise KeyError("Interface %s not initialized" % n)
        return self

    ###
    # ASSEMBLY PHASE
    ###

    def assemble(self):
        ### Override to combine components' drawings to final drawing
        pass

    def append(self, name, prefix, **kwargs):
        component = self.getSubcomponent(name)

        allPorts = set()
        for key in component.interfaces:
          try:
            allPorts.update(component.getInterface(key))
          except TypeError:
            # interface is not iterable, i.e. a single port
            allPorts.add(component.getInterface(key))
        for port in allPorts:
          port.prefix(prefix)

        for (key, composable) in component.composables.items():
            self.composables[key].append(composable, prefix, **kwargs)

    def attach(self, interface_1, interface_2, kwargs):
        (fromName, fromPort) = interface_1
        (toName, toPort) = interface_2
        for (key, composable) in self.composables.items():
            try:
                composable.attachInterfaces(self.getInterfaces(fromName, fromPort),
                                            self.getInterfaces(toName, toPort),
                                            kwargs)
            except:
                logstr  =  "Error in attach: \n"
                logstr += f"  from ({fromName}, {fromPort}): "
                logstr += self.getInterfaces(fromName, fromPort).toString()
                logstr +=  "\n"
                logstr += f"  to ({toName}, {toPort}): "
                logstr += self.getInterfaces(toName, toPort).toString()
                log.error(logstr)
                raise

    ###
    # BUILD PHASE
    ###

    def modifyParameters(self):
        # Override to manually specify how parameters get set during build
        pass

    def componentGraph(self):
        graph = nx.DiGraph()
        graph.graph['depth'] = 0
        for name, sc in self.subcomponents.items():
            scg = sc['object'].componentGraph()
            graph.add_node(name, name=name, component=sc['classname'], graph=scg)
            graph.graph['depth'] = max(graph.graph['depth'], scg.graph['depth']+1)
        for ((fromComponent, fromPort), (toComponent, toPort), kwargs) in list(self.connections.values()):
            graph.add_edge(fromComponent, toComponent)
        return graph

    def traverseGraph(self):
        graph = self.componentGraph()
        try:
            nx.find_cycle(graph, orientation='original')
        except nx.NetworkXNoCycle:
            pass
        else:
            log.warning("Cycle found in subcomponent connection graph, behavior is currently unspecified")
        return list(nx.topological_sort(graph))

    def evalConstraints(self, subComponent):
        for (parameterName, obj) in self.getSubParameters(subComponent).items():
            try:
                parent = obj.get("subcomponent")
                if not parent:
                    raise AttributeError
                else:
                    parent = self.getSubcomponent(parent)
            except AttributeError:
                parent = self

            try:
              x = YamlFunction(obj).eval(parent)
            except Exception as e:
              log.error(f"Error trying to evaluate constraints for {parameterName} on {obj}:")
              log.error(repr(e))
              raise

            if x is not None:
              self.setSubParameter(subComponent, parameterName, x)

    # Append composables from all known subcomponents
    # (including ones without explicitly defined connections)
    # set useDefaultParameters = False to replace unset parameters with sympy variables
    def evalComponent(self, name, useDefaultParameters=True):
        sc = self.subcomponents[name]
        obj = sc['object']
        classname = sc['classname']
        try: 
          kwargs = sc['kwargs']
        except IndexError:
          kwargs = {}

        try:
            obj.make(useDefaultParameters)
            for (key, composable) in obj.composables.items():
                if key not in self.composables:
                    self.composables[key] = composable.new()
            self.append(name, name, **kwargs)
        except:
            log.error("Error in subclass %s, instance %s" % (classname, name))
            raise

    def evalConnections(self, scName):
        for ((fromComponent, fromPort), (toComponent, toPort), kwargs) in self.connections.values():
            if toComponent != scName:
                continue
            for k, v in kwargs.items():
                kwargs[k] = YamlFunction(v).eval(self)
            self.attach((fromComponent, fromPort),
                        (toComponent, toPort),
                        kwargs)

    def informComposables(self):
      # Let composables know what components and interfaces exist
      # TODO remove this when we have a better way of letting composables
      # know about components that have no ports (ex Bluetooth module driver)
      for (key, composable) in self.composables.items():
        for (name, sc) in self.subcomponents.items():
          composable.addComponent(sc['object'])
        for (name, value) in self.interfaces.items():
          if value is not None:
            composable.addInterface(self.getInterface(name))

    # set useDefaultParameters = False to replace unset parameters with sympy variables
    def make(self, useDefaultParameters=True):
        if useDefaultParameters:
            self.useDefaultParameters()
        self.modifyParameters()

        scOrder = self.traverseGraph()
        for scName in scOrder:
            self.evalConstraints(scName)
            self.evalComponent(scName, useDefaultParameters)    # Merge composables from all subcomponents and tell them my components exist
            self.evalConnections(scName)   # Tell composables which interfaces are connected

        self.informComposables()    # Tell composables that my interfaces exist
        self.assemble()

    ###
    # OUTPUT PHASE
    ###

    def makeComponentMap(self, scd=None, full=None, idnum = 0):
        if scd is None:
            g = self.componentGraph()
            full = nx.DiGraph()
            full.add_node(0, name="", component=self.getName())
            title = f"<b>{self.getName()}</b>"
        else:
            g = scd['graph']
            title = f"<b>{scd['component']}</b><br>{scd['name']}"

        data = dict(id=idnum, 
                    datum=(g.graph['depth']+1)**2, 
                    children=[dict(id=len(full), datum=0.5)])
        full.add_node(len(full), title=title)

        if g:
            f = nx.convert_node_labels_to_integers(g, len(full))
            full = nx.disjoint_union(full, f)
            for sc, scd in f.nodes(data=True):
                full, cd = self.makeComponentMap(scd, full, sc)
                data["children"].append(cd)
        return full, data

    def drawComponentTree(self, basename, stub):
        g, data = self.makeComponentMap()
        circles = circlify([data])

        tree = gv(g,
            pos = {c.ex['id']: (c.x, c.y) for c in circles},
            node_size = {c.ex['id']: c.r for c in circles},
            node_text = lambda k, v : v.get("title", ""),
            node_border_width = lambda k, v: "title" not in v and 2 or 0,
            edge_width = 0,
        )
        fig = tree.create_figure(width=1024, height=768)

        if basename:
            fig.write_image(basename+stub)
            ret = basename+stub
        else:
            ret = fig.to_image(format="png")
        return ret

    def makeOutput(self, filedir=".", **kwargs):
        log.info(f"Compiling robot designs to directory {filedir} ...")
        def kw(arg, default=kwargs.get("default", True)):
            return kwargs.get(arg, default)

        if kw("remake", True):
            self.make(kw("useDefaultParameters", True))
            log.debug(f"... done making {self.getName()}.")

        # XXX: Is this the right way to do it?
        import os
        try:
            os.makedirs(filedir)
        except:
            pass

        # Process composables in some ordering based on type
        orderedTypes = ['electrical', 'ui', 'code'] # 'code' needs to know about pins chosen by 'electrical', and 'code' needs to know about IDs assigned by 'ui'
        # First call makeOutput on the ones of a type whose order is specified
        rets = {}
        for composableType in orderedTypes:
            if composableType in self.composables:
                kwargs["name"] = composableType
                ss = self.composables[composableType].makeOutput(filedir, **kwargs)
                rets[composableType] = ss
        # Now call makeOutput on the ones whose type did not care about order
        for (composableType, composable) in self.composables.items():
            if composableType not in orderedTypes:
                kwargs["name"] = composableType
                ss = self.composables[composableType].makeOutput(filedir, **kwargs)
                rets[composableType] = ss

        if kw("tree"):
            log.info("Generating hierarchy tree...")
            rets["component"] = self.drawComponentTree(filedir, "/tree.png")
            log.debug("done making tree.")

        log.info("Happy roboting!")
        return rets

    ###
    # OTHER STUFF
    # (probably obsolete)
    ###

    def addSemanticConstraint(self, lhs, op, rhs):
        self.semanticConstraints.append([lhs, op, rhs])

