from functools import reduce
from operator import add
import numpy
import sympy


# Common mods
def list_eye(x):
    return [[int(i==j) for j in range(x)] for i in range(x)]

def reduce_sum(a):
  return reduce(add, a)

def cumsum(iterable):
  arr = [iterable[0]]
  for c in iterable[1:]:
    arr.append(arr[-1] + c)
  return arr

# Numpy mods
def numpy_rows(x):
    return numpy.array(x)
  
def numpy_N(x):
    return x

def numpy_difference(pts1, pts2):
    return numpy.linalg.norm(numpy.array(pts1) - numpy.array(pts2))

def numpy_dex(pts1, pts2, tol):
    return numpy_difference(pts1, pts2) > tol

def numpy_pi():
    return numpy.pi

def numpy_dotrot(x, rot, angle):
    return numpy.dot(x, rot(numpy.deg2rad(angle)))

def numpy_det(m):
    return numpy.linalg.det(numpy.array(m))

# Sympy mods
def sympy_deg2rad(x):
    return x * (sympy.pi / 180)

def sympy_rad2deg(x):
    return x / (sympy.pi / 180)

def sympy_dot(a, b):
    return sympy.Matrix(a) * sympy.Matrix(b)

def sympy_norm(x):
    return sympy.Matrix(x).norm()

def sympy_inv(x):
    return x.inv()

def sympy_diag(x):
    return sympy.diag(*x)

def sympy_rows(x):
    return [x.row(i) for i in range(x.rows)]

def sympy_round(x):
    return x.round()

def sympy_difference(pts1, pts2):
    #XXX Hack to overcome precision errors
    from random import random
    pts1 = sympy.Matrix(pts1)
    pts2 = sympy.Matrix(pts2)

    syms = pts1.atoms(sympy.Symbol) | pts2.atoms(sympy.Symbol)
    subs = [(x, 100 + 100*random()) for x in syms]
    p1 = numpy.array(pts1.subs(subs)).astype(numpy.float64)
    p2 = numpy.array(pts2.subs(subs)).astype(numpy.float64)
    return numpy_difference(p1, p2)

def sympy_dex(pts1, pts2, tol):
    return sympy_difference(pts1, pts2) > tol

def sympy_pi():
    return sympy.pi

def sympy_dotrot(x, rot, angle):
    return x * rot(sympy_deg2rad(angle))

def sympy_det(m):
    return sympy.Matrix(m).det()

known_fns = {
  "cos"       : ( numpy.cos         , sympy.cos         ),
  "sin"       : ( numpy.sin         , sympy.sin         ),
  "tan"       : ( numpy.tan         , sympy.tan         ),
  "sqrt"      : ( numpy.sqrt        , sympy.sqrt        ),
  "transpose" : ( numpy.transpose   , sympy.transpose   ),
  "arctan2"   : ( numpy.arctan2     , sympy.atan2       ),
  "arccos"    : ( numpy.arccos      , sympy.acos        ),
  "array"     : ( numpy.array       , sympy.Matrix      ),
  "dot"       : ( numpy.dot         , sympy_dot         ),
  "norm"      : ( numpy.linalg.norm , sympy_norm        ),
  "inv"       : ( numpy.linalg.inv  , sympy_inv         ),
  "diag"      : ( numpy.diag        , sympy_diag        ),
  "deg2rad"   : ( numpy.deg2rad     , sympy_deg2rad     ),
  "rad2deg"   : ( numpy.rad2deg     , sympy_rad2deg     ),
  "round"     : ( numpy.round       , sympy_round       ),
  "N"         : ( numpy_N           , sympy.N           ),
  "rows"      : ( numpy_rows        , sympy_rows        ),
  "difference": ( numpy_difference  , sympy_difference  ),
  "differenceExceeds": ( numpy_dex  , sympy_dex         ),
  "pi"        : ( numpy_pi          , sympy_pi          ),
  "dotrot"    : ( numpy_dotrot      , sympy_dotrot      ),
  "eye"       : ( list_eye          , list_eye          ),
  "sum"       : ( reduce_sum        , reduce_sum        ),
  "cumsum"    : ( cumsum            , cumsum            ),
  "det"       : ( numpy_det         , sympy_det         ),
}

def __getattr__(fn):
    if fn not in known_fns:
        # raise AttributeError(f"{fn} not found in rocolib math library")
        return getattr(sympy, fn)

    ### TODO: Find a better way of determining whether any of the arguments are sympy expressions?
    def isSymbolic(a):
        return "sympy" in repr(type(a))
    def isSym(args):
        for a in args:
            if isSymbolic(a):
                return True
            if hasattr(a, '__iter__'):
                if isSym(a):
                    return True
        return False

    def choose(*args, **kwargs):
        if isSym(args):
            return known_fns[fn][1](*args, **kwargs)
        else:
            return known_fns[fn][0](*args, **kwargs)

    return choose
